$NOLIST

; Include Datei

;********************************************************************************************************
; Ansteuerung einer LCD-Anzeige (z.B. Typ 162) mit 1 - 4 Zeilen, 16 oder 24 Zeichen

; Displayspezfische Definitionen in Include-Datei:
; LCD16x - Display-162.a51
; LCD204 - Display-204.a51
; im Hauptprogramm muss z.B. stehen:
;   $INCLUDE(Display-162.a51)
;   $INCLUDE(Display.a51)

; J. Rathlev, Jun. 2006

; Bereitgestellte Unterroutinen:
;===============================
;   InitDisplay  - Anzeige initialisieren und lschen
;   DefChar      - Benutzerdefiniertes Zeichen speichern
;                  in DPTR: Zeiger auf Pixeltabelle (z.B. AuTab, AdTab)
;   WDChar       - ein Zeichen an der aktuellen Cursorpos. anzeigen
;                  in A: Zeichen
;   WDFStr       - festen nullterminierten String anzeigen 
;                  in DPTR: Adresse des ersten Zeichens im CodeMem
;   WDStr        - programmdefinierten nullterminierten String anzeigen 
;                  in DPTR: Adresse des ersten Zeichens im ExtMem
;   WDDec        - Byte als Dezimalzahl anzeigen (1. Null wird unterdrckt)
;                  in A: Wert der Zahl
;   SDLine       - Cursor an den Anfang einer Zeile setzen
;                  in A: Zeilennummer (0,1,,..)
;   SDLPos       - Cursor an beliebige Position einer Zeile setzen
;                  in A: Zeilennummer (0,1,,..)
;                     R2: Cursorposition
;   DCLeft       - Cursor um ein Zeichen nach links bewegen
;   DCRight      - Cursor um ein Zeichen nach rechts bewegen
;   DDLeft       - Gesamte Anzeige um ein Zeichen nach links schieben
;   DDRight      - Gesamte Anzeige um ein Zeichen nach rechts schieben
;   SDAPos       - Cursor an beliebige Position der aktuellen Zeile setzen
;                  in R2: Cursorposition                      
   
;********************************************************************************************************
; Konstanten
; ==========

LcdPort EQU     P1
LcdRS   EQU     10100000B       ; Signal RS
RSBit   EQU     7
LcdRW   EQU     01100000B       ; Signal RW
RWBit   EQU     6
LcdE    EQU     00110000B       ; Signal E
EBit    EQU     4
FourBit EQU     0010B           ; Funktion 4-Bit-Modus

;********************************************************************************************************

        DSEG

; RAM-Daten
; =========

Char:   DS      1               ; an Anzeige zu sendendes Zeichen
CCode:  DS      1               ; an Anzeige zu sendender Steurcode
Line:   DS      1               ; Anzeigezeile (0, 1)


;********************************************************************************************************
; Routinen fr die LCD-Anzeige
; ============================

        CSEG   

; Steuer-Byte schreiben
; Daten in A

WDCtrl: 
        MOV     CCode,A
        SWAP    A               ; zuerst Bit 4 .. 7
        ANL     A,#0FH
        ORL     A,#LcdE
        MOV     LcdPort,#LcdE   ; E = 1
        MOV     LcdPort,A       ; 4 Bit ausgeben
        NOP
        CLR     LcdPort.EBit    ; E = 0
        MOV     A,CCode
        ANL     A,#0FH          ; Bit 0 .. 3
        ORL     A,#LcdE
        SETB    LcdPort.EBit    ; E = 1
        MOV     LcdPort,A       ; 4 Bit ausgeben
        NOP
        CLR     LcdPort.EBit    ; E = 0
        RET

; Status-Byte lesen (Funktion ???)
; benutzt: R7
; Byte in A

RDStat: MOV     LcdPort,#LcdRW  ; auf Lesen     
        SETB    LcdPort.EBit    ; E = 1
        NOP
        CLR     LcdPort.EBit    ; E = 0
        NOP
        MOV     A,LcdPort       ; 4 Bit Daten lesen
        ANL     A,0FH
        SWAP    A
        MOV     R7,A
        SETB    LcdPort.EBit    ; E = 1
        NOP
        CLR     LcdPort.EBit    ; E = 0
        NOP
        MOV     A,LcdPort       ; 4 Bit Daten lesen
        ANL     A,0FH
        ORL     A,R7
        CLR     LcdPort.RWBit   ; RW = 0
        RET        


;********************************************************************************************************
; Warten bis Anzeige fertig ist
; kurze Warteschleife (ca. 100 s bei 24 MHz)

ShDel:  MOV     A,#100
ShDLp:  DEC     A
        JNZ     ShDLp
;        CALL    RStat          ; funktioniert nicht
;        JB      ACC.7,ShDel
        RET

; lange Warteschleife (ca. 2 ms bei 24 MHz)
; benutze R7

LgDel:  MOV     R7,#5
        CLR     A
LgDLp:  INC     A                              ; warten 
        JNZ     LgDLp
        DJNZ    R7,LgDLp
        RET


;********************************************************************************************************
; Sonderzeichen fr Anzeige definieren
AuTab:  DB     8,4,0EH,15H,4,4,4,4,0    ; Pfeil nach oben  (Char 0)
AdTab:  DB     16,4,4,4,4,15H,0EH,4,0    ; Pfeil nach unten (Char 1) 

; in DPTR Zeiger auf Tabelle (CharCode + 8 * Zeilenwerte)
; benutzt: R5,R6,R7

DefChar:
       MOV     R7,#8            ; Zhler fr Zeilen
       MOV     R6,#0
       CLR     A
       MOVC    A,@A+DPTR        ; Char Code * 8
       INC     DPTR
       ORL     A,#40H           ; CG RAM Address Set
       MOV     R5,A
DcLp:  MOV     A,R5
       ORL     A,R6
       CALL    WDCtrl
       MOV     A,R6       
       INC     R6
       MOVC    A,@A+DPTR
       CALL    WDChar     
       DJNZ    R7,DcLp
       RET       

;********************************************************************************************************
; Daten-Byte ausgeben
; Byte in A

WDChar:  
        MOV     Char,A
        CALL    ShDel
        MOV     LcdPort,#LcdRS  ; auf Daten     
        MOV     A,Char
        SWAP    A               ; zuerst Bit 4 .. 7
        ANL     A,#0FH
        ORL     A,#LcdE or LcdRS
        SETB    LcdPort.EBit    ; E = 1
        MOV     LcdPort,A       ; 4 Bit ausgeben
        NOP
        CLR     LcdPort.EBit    ; E = 0
        MOV     A,Char
        ANL     A,#0FH          ; Bit 0 .. 3
        ORL     A,#LcdE or LcdRS
        SETB    LcdPort.EBit    ; E = 1
        MOV     LcdPort,A       ; 4 Bit ausgeben
        NOP
        CLR     LcdPort.EBit    ; E = 0
        NOP
        CLR     LcdPort.RSBit   ; RS = 0        
        RET

;********************************************************************************************************
; Zeile/Position auswhlen
; Zeilennummer in A
; SDLPos: Zeilenposition in R2
; SDLine: an Zeilenanfang
; benutzt: R
 
SDLine: MOV     R2,#0
SDLPos: MOV     Line,A
SDAPos: PUSH    DP0H
        PUSH    DP0L
        CALL    ShDel
        MOV     A,Line
        ANL     A,#3
        MOV     DPTR,#DDAddr
        MOVC    A,@A+DPTR
        ADD     A,R2
        ORL     A,#80H           ; Adresse setzen
        CALL    WDCtrl
        CALL    ShDel
        MOV     A,#06H           ; AutoInc
        CALL    WDCtrl
        POP     DP0L     
        POP     DP0H
        RET

;********************************************************************************************************
; ausgewhlte Zeile lschen

CDLine:
        MOV     R7,#CharCount
CDLp:   MOV     A,#32         ; Leerzeichen
        CALL    WDChar
        DJNZ    R7,CDLp  
        MOV     A,Line
        CALL    SDLine              
        RET

;********************************************************************************************************
; Cursor um eine Position (links oder rechts) verschieben

DCLeft: 
        MOV     A,#10000B     ; links
        CALL    WDCtrl
        RET

DCRight: 
        MOV     A,#10100B     ; rechts
        CALL    WDCtrl
        RET
        
;********************************************************************************************************
; Anzeige um eine Position (links oder rechts) verschieben

DDLeft: 
        MOV     A,#11000B     ; links
        CALL    WDCtrl
        RET

DDRight: 
        MOV     A,#11100B     ; rechts
        CALL    WDCtrl
        RET
        
        
;********************************************************************************************************
; letztes Zeichen lschen
DDel:
        CALL    DCLeft
        MOV     A,#32
        CALL    WdChar
        CALL    ShDel
        CALL    DCLeft
        RET
           
;********************************************************************************************************
; Nulltem. String aus Code-Mem. schreiben
; Zeiger auf String in DPTR

WDFStr:
        CLR     A 
        MOVC    A,@A+DPTR
        INC     DPTR
        JZ      WDEnd
        CALL    WDChar
        SJMP    WDFStr
        
WDEnd:  RET

;********************************************************************************************************
; Nulltem. String aus Code-Mem. schreiben
; Zeiger auf String in DPTR

WDStr:
        MOVX    A,@DPTR
        INC     DPTR
        JZ      WDEnd
        CALL    WDChar
        SJMP    WDStr
        
;********************************************************************************************************
; Byte als Dezimalzahl anzeigen
; 1. Null wird unterdrckt
; Wert in A

WDDec:
        MOV     B,#100
        DIV     AB              ; 1. Stelle in A, Rest in B
        JZ      WDD1
        ADD     A,#30H          ; ASCII-Code
        CALL    WDCHAR
WDD1:   MOV     A,B
        MOV     B,#10
        DIV     AB              ; 2. Stelle
        ADD     A,#30H          ; ASCII-Code
        CALL    WDCHAR
        MOV     A,B             ; 3. Stelle
        ADD     A,#30H          ; ASCII-Code
        CALL    WDCHAR
        RET

ClearDisplay:
        CALL    ShDel
        MOV     A,#01H                        ; Display lschen, Cursor an Anfang
        CALL    WDCtrl
        RET

;********************************************************************************************************
; Anzeige im 4-Bit-Modus initialisieren

InitDisplay:
        MOV     LcdPort,#0
        MOV     LcdPort,#LcdE   ; E = 1
        MOV     LcdPort,#LcdE or FourBit       ; auf 4-Bit-Modus umschalten
        CALL    ShDel
        CLR     LcdPort.EBit                   ; E = 0
        CALL    LgDel

        MOV     A,#FourBit shl 4 or 8          ; 2-Zeilen, 5x7
        CALL    WDCtrl

        CALL    ShDel
        CALL    ShDel
        MOV     A,#0FH                        ; Display ein, Cursor ein, Cursor blinken
        CALL    WDCtrl

        CALL    LgDel
        MOV     A,#01H                        ; Display lschen, Cursor an Anfang
        CALL    WDCtrl

        CALL    LgDel
        MOV     A,#06H                        ; Cursor auf "Autoinkrement"
        CALL    WDCtrl
        CALL    ShDel
        RET        

$LIST
