$NOLIST

;********************************************************************************************************
; Include Datei
; Treiberpaket fr I2C-Bus
; J. Rathlev, IEAP, Uni-Kiel, Jan. 2008

; Portanschlsse
 
SSCL       EQU     P3.5
SSDA       EQU     P3.4


; EEPROMs  (24C04,..)

EPRADR     EQU     0A0H         ; I2C-Adresse des EEPROM
EPROFS     EQU     0            ; Adresse im EEPROM fr WS-Status

; Port-Expander PCF8574

PE0ADR     EQU     040H         ; I2C-Basisadresse des PCF8574
PE1ADR     EQU     070H         ; I2C-Basisadresse des PCF8574A
         
; AD/DA-Wandler PCF8591

ADADR      EQU     090H         ; I2C-Basisadresse des PCF8591

; Temperatursensor LM 75

TMADR      EQU     090H         ; I2C-Basisadresse des LM 75

;********************************************************************************************************

        CSEG

; Lesen und Schreiben von Bytes ber den I2C-Bus
; nach ELVjournal 5/96

; Startbedingung erzeugen
; benutzt :     R7
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

I2CStart:
        SETB    SSCL            ; SCL = H
        NOP
        SETB    SSDA            ; SDA = H
        NOP
        MOV     R7,#10          ; 20s prfen und warten
        SETB    C               ; Fehlerflag setzen
I2CSTW: JNB     SSCL,I2CSTE     ; Fehler: SCL = L (2 s)
        JNB     SSDA,I2CSTE     ; Fehler: SDA = L (2 s)
        DJNZ    R7,I2CSTW       ; warten          (2 s)
        CLR     SSDA            ; SDA = L
        NOP
        CLR     SSCL            ; SCL = L
        CLR     C               ; ok
I2CSTE: RET


; Stoppbedingung erzeugen

I2CStop:
        CLR     SSDA            ; SDA = L
        NOP
        SETB    SSCL            ; SCL = H
        NOP
        SETB    SSDA            ; SDA = H
        RET


; ein Bit auf Bus schreiben
; Eingabe:      C = Datenbit
; benutzt:      R7
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

I2CWBIT:
        MOV     SSDA,C          ; Bit ausgeben
        NOP
        JNC     I2CW1           ; Bit = L ?
        MOV     R7,#10          ; ja - prfen, ob Bus frei
I2CWW1: JB      SSDA,I2CW1      ; SDA = H ?
        DJNZ    R7,I2CWW1       ; nein, warten
        SETB    SSCL            ; Bus belegt, freigeben
        SETB    C
        SJMP    I2CWE           ; Fehler, C = 1

I2CW1:  SETB    SSCL            ; SCL = H
        NOP
        MOV     R7,#10          ; ja - prfen, ob Bus frei
I2CWW2: JB      SSCL,I2CW2      ; SCL = H ?
        DJNZ    R7,I2CWW2       ; nein, warten
        SETB    SSDA            ; Bus belegt, freigeben
        SETB    C
        SJMP    I2CWE           ; Fehler, C = 1

I2CW2:  CLR     SSCL            ; SCL = L
        CLR     C               ; C = 0 - ok
I2CWE:  RET


; ein Bit von Bus lesen
; benutzt:      R7
; Ergebnis:     F0 = Datenbit
;               C = 1 : Fehler
;               C = 0 : ok

I2CRBIT:
        SETB    SSDA            ; SDA = H
        NOP
        SETB    SSCL            ; SCL = H
        NOP
        MOV     R7,#10          ; prfen, ob Bus frei
I2CRW1: JB      SSCL,I2CR1      ; SCL = H ?
        DJNZ    R7,I2CRW1       ; nein, warten
        SETB    C               ;
        SJMP    I2CRE           ; Fehler, C = 1

I2CR1:  MOV     C,SSDA          ; lese Bit
        MOV     F0,C
        CLR     SSCL            ; SCL = L
        CLR     C               ; C = 0 - ok
I2CRE:  RET


; ein Byte ausgeben
; Eingabe:      A = Datenbyte
; benutzt:      R6,R7,F0
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

I2CWBYTE:
        MOV     R6,#8           ; Zhler
I2CWB1: RLC     A               ; Bits nach C
        ACALL   I2CWBIT         ; Bit schreiben
        JC      I2CWBE          ; Fehler
        DJNZ    R6,I2CWB1       ; Schleife fr 8 Bits
        ACALL   I2CRBIT         ; Besttigungsbit (F0) lesen
        JC      I2CWBE          ; Fehler
        MOV     C,F0            ; Besttigungsbit prfen
        JNC     I2CWBE          ; ACK = L
        ACALL   I2CSTOP         ; nein, Stop
I2CWBE: RET


; ein Byte lesen
; Eingabe:      C = L : mit ACK im 9. Bit
; benutzt:      R5,R6,R7,F0,A
; Ergebnis:     A = Datenbyte
;               C = 1 : Fehler
;               C = 0 : ok

I2CRBYTE:
        RLC     A               ; C nach LSB
        MOV     R5,A            ; zwischenspeichern
        MOV     R6,#8           ; Zhler
I2CRB1: ACALL   I2CRBIT         ; Bit lesen
        JC      I2CRBE          ; Fehler
        MOV     C,F0
        RLC     A               ; Bit nach A
        DJNZ    R6,I2CRB1       ; Schleife fr 8 Bits
        XCH     A,R5
        RRC     A               ; LSB nach C
        XCH     A,R5
        ACALL   I2CWBIT         ; Besttigungsbit setzen
        SETB    SSDA            ; SDA = H
I2CRBE: RET



;********************************************************************************************************
; mehrere Bytes in das EEPROM (Bank wie in EPRADR) schreiben
; Eingabe:      R2 = Adresse im EEPROM
;               R1 = Anzahl der Bytes (max. 8 modulo Adresse)
;               R0 = Zeiger auf Bytes im RAM
; benutzt:      R6,R7,F0,A
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

EPWBlk:
        MOV     A,R1            ; Teste Anzahl
        JZ      EPWST           ; Ende
        ACALL   I2CSTART        ; Start
        JC      EPWBLE          ; Fehler
        MOV     A,#EPRADR       ; Baustein-Adresse und Bank
        CLR     ACC.0           ; Bit 0 lschen
        ACALL   I2CWBYTE        ; schreiben
        JC      EPWBLE          ; Fehler
        MOV     A,R2            ; Adresse
        ACALL   I2CWBYTE        ; schreiben
        JC      EPWBLE          ; Fehler
EPWLP:  MOV     A,@R0           ; Datenbyte
        ACALL   I2CWBYTE        ; schreiben
        JC      EPWBLE          ; Fehler
        INC     R0
        DJNZ    R1,EPWLP        ; wiederholen
EPWST:  ACALL   I2CSTOP         ; Ende
EPWBLE: RET


; mehrere Bytes aus dem EEPROM (Bank wie in EPRADR) lesen
; Eingabe:      R2 = Adresse im EEPROM
;               R1 = Anzahl der Bytes (max. 8 modulo Adresse)
;               R0 = Zeiger auf Bytes im RAM
; benutzt:      R5,R6,R7,F0
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

EPRBlk:
        MOV     A,R1            ; Teste Anzahl
        JZ      EPRST           ; Ende
        ACALL   I2CSTART        ; Start
        JC      EPRBLE          ; Fehler
        MOV     A,#EPRADR       ; Baustein-Adresse und Bank
        CLR     ACC.0           ; Bit 0 lschen
        ACALL   I2CWBYTE        ; schreiben
        JC      EPRBLE          ; Fehler
        MOV     A,R2            ; Adresse
        ACALL   I2CWBYTE        ; schreiben
        JC      EPRBLE          ; Fehler
        ACALL   I2CSTART        ; neue Startbedingung erzeugen
        JC      EPRBLE          ; Fehler
        MOV     A,#EPRADR       ; Baustein-Adresse und Bank
        SETB    ACC.0           ; Bit 0 setzen
        ACALL   I2CWBYTE        ; schreiben
        JC      EPRBLE          ; Fehler
        DEC     R1
        CJNE    R1,#0,EPRLP     ; mehr als 1 Byte
        SJMP    EPRLB           ; nur 1 byte
EPRLP:  ACALL   I2CRBYTE        ; lesen
        JC      EPRBLE          ; Fehler
        MOV     @R0,A           ; Datenbyte speichern
        INC     R0
        DJNZ    R1,EPRLP        ; wiederholen
EPRLB:  SETB    C               ; kein ACK fr letztes Byte
        ACALL   I2CRBYTE        ; letztes Byte lesen
        JC      EPRBLE          ; Fehler
        MOV     @R0,A           ; Datenbyte speichern
EPRST:  ACALL   I2CSTOP         ; Ende
EPRBLE: RET



;********************************************************************************************************
; ein oder mehrere Bytes in den Port-Expander schreiben
; Eingabe:      R2 = Unter-Adresse des Port-Expanders (0 .. 7)
;               R1 = Anzahl der Bytes
;               R0 = Zeiger auf Byte im RAM
; benutzt:      R6,R7,F0,A,B
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

PE1Write:
        MOV     B,#PE1ADR
        JMP     PEWADR
PE0Write:
        MOV     B,#PE0ADR
PEWADR: MOV     A,R2            ; I2C-Adresse erzeugen
        ANL     A,#7
        CLR     C
        RLC     A               ; ein Bit nach links
        ADD     A,B
        ACALL   I2CSTART        ; Start
        JC      PEWERR          ; Fehler
        ACALL   I2CWBYTE        ; Adresse schreiben
        JC      PEWERR          ; Fehler
PEWLP:  MOV     A,@R0           ; Datenbyte
        ACALL   I2CWBYTE        ; schreiben
        JC      PEWERR          ; Fehler
        INC     R0
        DJNZ    R1,PEWLP        ; wiederholen
        ACALL   I2CSTOP         ; Ende
PEWERR: RET


; ein oder mehrere Bytes aus dem Port-Expander lesen
; Eingabe:      R2 = Unter-Adresse des Port-Expanders (0 .. 7)
;               R1 = Anzahl der Bytes (max. 8 modulo Adresse)
;               R0 = Zeiger auf Bytes im RAM
; benutzt:      R5,R6,R7,F0,A,B
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

PE1Read:
        MOV     B,#PE1ADR
        JMP     PERADR
PE0Read:
        MOV     B,#PE0ADR
PERADR: MOV     A,R2            ; I2C-Adresse erzeugen
        ANL     A,#7
        SETB    C
        RLC     A               ; ein Bit nach links
        ADD     A,B
        ACALL   I2CSTART        ; Start
        JC      PERERR          ; Fehler
        ACALL   I2CWBYTE        ; Adresse schreiben
        JC      PERERR          ; Fehler
        DEC     R1
        CJNE    R1,#0,PERLP     ; mehr als 1 Byte
        SJMP    PERLB           ; nur 1 Byte
PERLP:  ACALL   I2CRBYTE        ; lesen
        JC      PERERR          ; Fehler
        MOV     @R0,A           ; Datenbyte speichern
        INC     R0
        DJNZ    R1,PERLP        ; wiederholen
PERLB:  SETB    C               ; kein ACK fr letztes Byte
        ACALL   I2CRBYTE        ; letztes Byte lesen
        JC      PERERR          ; Fehler
        MOV     @R0,A           ; Datenbyte speichern
        INC     R0
        ACALL   I2CSTOP         ; Ende
PERERR: RET

;********************************************************************************************************
; ein oder mehrere Bytes in den AD/DA-Wandler PCF8591 schreiben
; Eingabe:      R2 = Unter-Adresse des Sensors (0 .. 7)
;               R1 = Anzahl der Bytes
;               R0 = Zeiger auf Byte im RAM
; benutzt:      R6,R7,F0,A,B
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

AdWrite:
        MOV     B,#AdAdr
        JMP     PEWADR

; ein oder mehrere Bytes aus dem AD/DA-Wandler PCF8591 lesen
; Eingabe:      R2 = Unter-Adresse des Sensors (0 .. 7)
;               R1 = Anzahl der Bytes (max. 8 modulo Adresse)
;               R0 = Zeiger auf Bytes im RAM
; benutzt:      R5,R6,R7,F0,A,B
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

AdRead:
        MOV     B,#AdAdr
        JMP     PERADR

;********************************************************************************************************
; ein oder mehrere Bytes in den Temperatur-Sensor LM75 schreiben
; Eingabe:      R2 = Unter-Adresse des Sensors (0 .. 7)
;               R1 = Anzahl der Bytes
;               R0 = Zeiger auf Byte im RAM
; benutzt:      R6,R7,F0,A,B
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

TmWrite:
        MOV     B,#TmAdr
        JMP     PEWADR

; ein oder mehrere Bytes aus dem Temperatur-Sensor LM75 lesen
; Eingabe:      R2 = Unter-Adresse des Sensors (0 .. 7)
;               R1 = Anzahl der Bytes (max. 8 modulo Adresse)
;               R0 = Zeiger auf Bytes im RAM
; benutzt:      R5,R6,R7,F0,A,B
; Ergebnis:     C = 1 : Fehler
;               C = 0 : ok

TmRead:
        MOV     B,#TmAdr
        JMP     PERADR

;********************************************************************************************************
$LIST
