$NOLIST

; Include Datei: SerInOut.a51

;********************************************************************************************************
; Datentransfer ber die serielle Schnittstelle
; Lesen von Bytes mit Interrupt und Zwischenpuffer
; Schreiben von Bytes ohne Interrupt
; bentigt externen Speicher
;          =================
; J. Rathlev, Juni 2006
;********************************************************************************************************

LF      EQU     0AH
CR      EQU     0DH

DPL     EQU     DP0L
DPH     EQU     DP0H

; int. Datenspeicher
; ==================

        DSEG    

; Buffer-Steuerung fr den interrupt-gepufferten Empfang
; von Daten ber die serielle Schnittstelle

ICCNT:  DS      2       ; Zeichenzhler
ILCNT:  DS      2       ; Zeilenzhler
IWPNT:  DS      2       ; Zeiger fr Schreiben in den Buffer
IRPNT:  DS      2       ; Zeiger fr Lesen aus dem Buffer

        BSEG

ORDY:   DBIT    1        
ZF:     DBIT    1        

;********************************************************************************************************
; ext. Datenspeicher
; ==================

        XSEG    

BUFL    EQU     0200H   ; Input-Bufferlnge

INBUF:  DS      BUFL    ; Input-Buffer

BASEAD  EQU     INBUF           ; Anfangsadresse des Buffers
ENDAD   EQU     BASEAD+BUFL     ; Endadresse des Buffers
 
;********************************************************************************************************
; Interrupt-Routine fr ser. Schnittstelle

        CSEG

SVPC    SET     $

        ORG     CodeMem+23H

        LJMP    SRINT           ; Interrupt von ser. Schnitstelle

        ORG     SVPC
        
USING   2

SRINT:  PUSH    PSW
        MOV     PSW,#010H       ; Registerbank 2
        PUSH    ACC             ; benutzte Register sichern
        PUSH    DPL
        PUSH    DPH
        JNB     RI,SSEND       ; kein RI-Interrupt

        MOV     A,SBUF          ; Zeichen lesen
        CLR     RI              ; Flag zurcksetzen
        MOV     R7,A            ; zwischenspeichern

        XRL     A,#LF           ; auf Line Feed prfen
        JZ      SRRET           ; ja - ignorieren

        MOV     A,ICCNT         ; Bufferzhler
        CLR     C
        SUBB    A,#LOW(BUFL)    ; <= BufL (low) ?
        MOV     A,ICCNT+1
        SUBB    A,#HIGH(BUFL)   ; <= BufL (high) ?
        JNC     SRRET           ; Buffer voll

        MOV     A,R7            ; Zeichen testen
        XRL     A,#CR           ; CR
        JNZ     SRJA            ; nein

SRJC:   INC     ILCNT           ; Zeilenzhler (low) erhhen
        MOV     A,ILCNT
        JNZ     SRJA
        INC     ILCNT+1         ; Zeilenzhler (high)

SRJA:   MOV     DPH,IWPNT+1     ; Bufferzeiger
        MOV     DPL,IWPNT
        MOV     A,R7
        MOVX    @DPTR,A         ; Zeichen in Buffer schreiben
        INC     DPTR            ; erhhen
        
        MOV     A,DPH
        XRL     A,#high(ENDAD)  ; Endadresse erreicht?
        JNZ     SRJ8
        MOV     A,DPL
        XRL     A,#low(ENDAD)
        JNZ     SRJ8
        MOV     DPH,#high(BASEAD)   ; auf Basisadresse setzen
        MOV     DPL,#low(BASEAD)
SRJ8:   MOV     IWPNT,DPL
        MOV     IWPNT+1,DPH

        INC     ICCNT           ; Bufferzhler (low)
        MOV     A,ICCNT
        JNZ     SRRET
        INC     ICCNT+1

SSEND:  JNB     TI,SRRET        ; TI-Interrupt?
        CLR     TI              ; ja
        SETB    ORDY
SRRET:  POP     DPH
        POP     DPL
        POP     ACC
        POP     PSW
        RETI

USING   0

;********************************************************************************************************
; Baudrateneinstellungen:
; SMOD = 1 ==> TH1 = 256 - (2 * f/Hz) / (384 * BR)

; 22,1184 MHz -   2400 : 208     24 MHz - 2400 : 204        12 MHz - 2400 : 230
;             -   4800 : 232            - 4800 : 230               - 4800 : 243
;             -   9600 : 244            - 9600 : 243
;             -  19200 : 250
;             -  38400 : 253
;             -  57600 : 254
;             - 115200 : 255

;********************************************************************************************************
; Serielle Schnittstelle initalisieren
; A     : Zhler fr Baudrate (s.o.)

InitSerial:   
        MOV     TH1,A           ; load value

        MOV     A,TMOD
        ANL     A,#0FH
        ADD     A,#20H
        MOV     TMOD,A          ; timer 1 mode 2 (8 bit, auto-reload)
        CLR     SM0             ; serial mode 1 (8-bit)
        SETB    SM1
        SETB    TR1             ; start timer
        MOV     A,PCON
        ORL     A,#80H
        MOV     PCON,A          ; SMOD=1 (Baudrate x 2)
        SETB    REN             ; Enable receiver
        CLR     TI              
        SETB    ORDY            ; set for transmit buffer empty
        CLR     RI              ; clear SI for receive interrupt
        CLR     ES
        RET

;  Buffer fr Datenempfang initialisieren, Interrupt einschalten

InitSerInput:
        CLR     A
        MOV     ICCNT,A         ;Zeichenzhler = 0
        MOV     ICCNT+1,A
        MOV     ILCNT,A         ;Zeilenzhler = 0
        MOV     ILCNT+1,A
        MOV     IWPNT,#low(INBUF)       ;Schreib- und Lesezeiger initialisieren
        MOV     IWPNT+1,#high(INBUF)
        MOV     IRPNT,#low(INBUF)
        MOV     IRPNT+1,#high(INBUF)

        SETB    ES              ; Serial-Interrupt freigeben
        RET

;********************************************************************************************************
; Zeichen von Serial-Port aus dem Zwischenpuffer holen
; benutzt: R7
; Zeichen in A und R7
; Ergebnis: C = 0 = o.k.
;           C = 1 = keine Daten

READCHAR:
        MOV     A,ICCNT         ; Bufferzhler (low)
        JNZ     RCH1
        MOV     A,ICCNT+1       ; Bufferzhler (high)
        JZ      RCH4            ; kein Zeichen im Buffer

RCH1:   DEC     ICCNT           ; Bufferzhler (low)
        MOV     A,ICCNT
        CJNE    A,#0FFH,RCH2
        DEC     ICCNT+1         ; Bufferzhler (high)

RCH2:   PUSH    DPL
        PUSH    DPH
        MOV     DPH,IRPNT+1     ; Bufferzeiger
        MOV     DPL,IRPNT
        MOVX    A,@DPTR         ; Zeichen aus Buffer
        MOV     R7,A            ; zwischenspeichern
        INC     DPTR            ; erhhen
        
        MOV     A,DPH
        XRL     A,#high(ENDAD)  ; Endadresse erreicht?
        JNZ     RCH3
        MOV     A,DPL
        XRL     A,#low(ENDAD)
        JNZ     RCH3    
        MOV     DPH,#high(BASEAD)   ; auf Basisadresse setzen
        MOV     DPL,#low(BASEAD)
RCH3:   MOV     IRPNT,DPL
        MOV     IRPNT+1,DPH
        POP     DPH
        POP     DPL

        MOV     A,R7            ; Zeichen nach A        
        CLR     C
        RET

RCH4:   SETB    C
        RET

;********************************************************************************************************
; Zeile von Serial-Port aus dem Zwischenpuffer holen
; benutzt: R6, R7
; max. Stringlnge in A
; Zeiger auf nullterm. String in DPTR (reservierter Platz: (A)+1)
; Ergebnis: C = 0 ==> o.k.
;           C = 1 ==> keine Daten

READLINE:
        JZ      RDS5            ; ungltige Stringlnge
        MOV     R6,A            ; max. Stringlnge
        MOV     A,ILCNT         ; Zeilenzhler (low)
        ORL     A,ILCNT+1       ; Zeilenzhler (high)
        JZ      RDS5            ; keine Zeilen im Buffer

RDS1:   DEC     ILCNT           ; Zeilenzhler (low)
        MOV     A,ILCNT
        CJNE    A,#0FFH,RDS2
        DEC     ILCNT+1         ; Zeilenzhler (high)

RDS2:   PUSH    DPL
        PUSH    DPH

RDLOOP: ACALL   READCHAR
        CLR     C
        SUBB    A,#020H         ; gltiges Zeichen?
        JNC     RDS3            ; ja

        MOV     A,R7
        XRL     A,#CR           ; Zeilenende?
        JZ      RDS4            ; ja

RDS3:   MOV     A,R6            ; Zhler
        JZ      RDLOOP          ; berlesen bis Zeilenende

        DEC     R6
        MOV     A,R7
        MOVX    @DPTR,A         ; speichern
        INC     DPTR    
        JMP     RDLOOP          ; nchstes Zeichen

RDS4:   CLR     A
        MOVX    @DPTR,A         ; Stringende
        POP     DPH
        POP     DPL

        CLR     C
        RET

RDS5:   SETB    C
        RET


;********************************************************************************************************
; Ein Zeichen an Serial-Port ausgeben
; Zeichen in A

WRITECHAR:
WWAIT:  JNB     ORDY,WWAIT      ; warten
        MOV     SBUF,A          ; Ausgabe
        CLR     ORDY
WRET:   RET


;********************************************************************************************************
; Nullterm. String in Daten-Speicher an Serial-Port ausgeben
; Zeiger auf String in DPTR

WRITESTRING:
        MOVX    A,@DPTR
        INC     DPTR
        JZ      WRET
        ACALL   WRITECHAR
        SJMP    WRITESTRING

;********************************************************************************************************
; Nullterm. String in Daten-Speicher an Serial-Port ausgeben
; und Cr-Lf anhngen
; Zeiger auf String in DPTR

WRITELINE:
        MOVX    A,@DPTR
        INC     DPTR
        JZ      WRITECR
        ACALL   WRITECHAR
        SJMP    WRITELINE
WRITECR:         
        MOV     A,#CR
        ACALL   WRITECHAR
        MOV     A,#LF
        ACALL   WRITECHAR
        RET


;********************************************************************************************************
; Nulltem. String aus Code-Mem. an Serial-Port ausgeben
; Zeiger auf String in DPTR

WRITEFIXSTR:
        CLR     A 
        MOVC    A,@A+DPTR
        INC     DPTR
        JZ      WFEND
        ACALL   WRITECHAR
        SJMP    WRITEFIXSTR
        
WFEND:  RET

;********************************************************************************************************
; Byte als Dezimalzahl an Serial-Port ausgeben
; Wert in A
; Carry = 0 : mit fhrenden Nullen
;       = 1 : ohne fhrende Nullen

WRITEDEC:
        MOV     ZF,C            ; speicher Flag
        MOV     B,#100
        DIV     AB              ; 1. Stelle in A, Rest in B
        JZ      WD3
        CLR     ZF
        JMP     WD2             ; nicht Null
        
WD3:    JNB     ZF,WD2          ; mit fhrender 0
        JZ      WD1
WD2:    ADD     A,#30H          ; ASCII-Code
        CALL    WRITECHAR
WD1:    MOV     A,B
        MOV     B,#10
        DIV     AB              ; 2. Stelle
        JZ      WD6
        CLR     ZF
        JMP     WD5             ; nicht Null

WD6:    JNB     ZF,WD5          ; mit fhrender 0
        JZ      WD4
WD5:    ADD     A,#30H          ; ASCII-Code
        CALL    WRITECHAR
WD4:    MOV     A,B             ; 3. Stelle
        ADD     A,#30H          ; ASCII-Code
        CALL    WRITECHAR
        RET

$LIST
